/*
   26 July 1998 Jonathan Chin
   Kibologised March 1999
*/

#define BEABLE_DRIVER
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include "beable.h"

struct strdata { const char *data; size_t size; };

static struct strdata str_beableones = {
	"\377",
	sizeof("\377")-1
};

static struct strdata str_beablezorro = {
	"zorro\n",
	sizeof("zorro\n")-1
};

static struct strdata str_beablebeable = {
	"beable\n",
	sizeof("beable\n")-1
};

static struct strdata str_beabledoidy = {
	"doidy\n",
	sizeof("doidy\n")-1
};

static struct file_operations beable_fops = {

	NULL,	/* lseek */
	beable_read,	/* read */
	beable_write,	/* write */
	NULL,	/* readdir */
	NULL,	/* select */
	beable_ioctl,	/* ioctl */
	NULL,	/* mmap */
	beable_open,	/* open */
	beable_release,	/* release */
	NULL,	/* fsync */
	NULL,	/* fasync */
	NULL,	/* check_media_change */
	NULL	/* revalidate */

};

static int n=0;
static int inputsize=0;
static int beable_active=0;
static struct strdata *beable_strdata=&str_beableones;

int init_module(void) {

	
	if (register_chrdev(BEABLE_MAJOR,BEABLE_NAME,&beable_fops)) {
		printk("Beable: Couldn't get major %d\n",BEABLE_MAJOR);
		return -EIO;
	}

	printk("Beable driver registered.\n");
	return 0;

}

void cleanup_module(void) {

	unregister_chrdev(BEABLE_MAJOR,BEABLE_NAME);
	printk("Beable driver unregistered.\n");
}


/* Called when a file descriptor to /dev/beable is opened */
/* Returns 0 for OK, else -ve error number */

static int beable_open(struct inode *inode, struct file *file)
{

	if (beable_active)
		return -EBUSY;
	beable_active=1;

	return 0;
}

/* Called when fd to /dev/beable is closed */

static void beable_release(struct inode *inode, struct file *file)
{
	if (!beable_active)
		printk("WTF?!? Trying to close unopened beable.\n");

	beable_active=0;
}

/* Read count bytes into buffer */
/* Return value is either no of bytes read, or -ve error */

static int beable_read(struct inode *indoe, struct file *file,
			char *buffer,int count)
{

	char *p=buffer;
	const char *data=beable_strdata->data;
	int i;

	int size=beable_strdata->size;

	if (count<size) {
		if (count<=(size-n)) {
			memcpy_tofs(p,data+n,count);
			n=(n+count)%size;
			return count;
		} else {
			memcpy_tofs(p,data+n,size-n);
			memcpy_tofs(p+size-n,data,count-size+n);
			n=count-size+n;
			return count;
		}

	} else {
		memcpy_tofs(p,data+n,size-n);
		p=p+size-n;
		for (i=0;i<((count-size+n)/size);i++) {
			memcpy_tofs(p,data,size);
			p=p+size;
		}
		memcpy_tofs(p,data,(count-size+n)%size);
		n=(count-size+n)%size;
		return count;	
	}
	printk("My brane is on drugs! \n");
	return -EINVAL;
}

/* Write count bytes into buffer */
/* Return either no of bytes written, or -ve error */

static int beable_write(struct inode *inode, struct file *file,
			const char *buffer,int count)
{
#ifdef	FINITE_BEABLE
	if (inputsize>BEABLE_MAX_INPUT)
		return -ENOSPC;
	
	if ((inputsize+count)>BEABLE_MAX_INPUT) {
		inputsize=BEABLE_MAX_INPUT+1;
		return -ENOSPC;
	}
#endif
	inputsize+=count;
	return count;
}

/* IOCTL of type cmd with arguments arg */
/* Arg is usually cast into a pointer.. */

static int beable_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	switch (cmd) {

		case BEABLEONES:
			beable_strdata=&str_beableones;
			n=0;
			break;

		case BEABLEZORRO:
			beable_strdata=&str_beablezorro;
			n=0;
			break;

		case BEABLEBEABLE:
			beable_strdata=&str_beablebeable;
			n=0;
			break;

		case BEABLEDOIDY:
			beable_strdata=&str_beabledoidy;
			n=0;
			break;

		case BEABLESWALLOW:
			inputsize=0;
			break;

		default:
			return -EINVAL;
	}
	return 0;
}
